#include "CubicSpline.h"

void CalculateDerivatives(MList<float> &,
         MList<float> &, MList<float> &,
         MList<float> &);

void CalculateDerivatives(MList<float> &points,
         MList<float> &first, MList<float> &second,
         MList<float> &third) {
   int i;
   int numKnots;
   float tmpFloat;
   float calc;
   MList<float> deltas;
   MList<float> tmpA;
   MList<float> tmpB;
   MList<float> tmpDerivs;

   numKnots = points.Len();

   // calculate delta^2
   for(i = 0 ; i < numKnots - 2 ; i++) {
      tmpFloat = points[i+2] - 
                  (2 * points[i+1]) +
                  points[i];
      deltas.Append(tmpFloat);
   }

   // set up for some intermediate calculations
   tmpA.Append(4.0);
   tmpB.Append(0.0);
   
   // do some intermediate calculations
   for(i = 1 ; i < numKnots - 2 ; i++) {
      tmpFloat = 4 - (1 / tmpA[i-1]);
      tmpA.Append(tmpFloat);
      
      tmpFloat = deltas[i] -
              (tmpB[i-1] / tmpA[i-1]);
      tmpB.Append(tmpFloat);
   }
   
   // calculate the second derivitaves in reverse order
   // the first and last second derivatives are 0
   tmpDerivs.Append(0.0);
   
   // calculate the rest of them
   for(i = numKnots - 1 ; i >= 2 ; i--) {
      tmpFloat = (6 * (tmpB[i-2]) -
                   tmpDerivs[numKnots - (i + 1)]) /
                    tmpA[i - 2];
      tmpDerivs.Append(tmpFloat);
   }
   
   // the first and last second derivatives are 0
   tmpDerivs.Append(0.0);

   // reverse the second derivitaves for pass back
   for(i = tmpDerivs.Len() - 1; i >= 0  ; i--) {
      tmpFloat = tmpDerivs[i];
      second.Append(tmpFloat);
   }
   
   // calculate first and third derivitaves
   for( i = 0; i < numKnots - 1; i++) {
      tmpFloat = points[i+1];
      tmpFloat -= points[i];
      calc = 2 * second[i];
      calc += second[i+1];
      calc /= 6.0;
      tmpFloat -= calc;
      first.Append(tmpFloat);
      
      tmpFloat = second[i+1] - second[i];
      third.Append(tmpFloat);
   }
}

CubicSpline::CubicSpline(List<Pt3d> &knots) {
   MList<float> xValues;
   MList<float> yValues;
   MList<float> zValues;

   float tmpFloat;
   int i;
   
   // build the parameterised Lists
   for(i=0; i < knots.Len(); i++) {
      tmpFloat = (knots[i])->GetX();
      xValues.Append(tmpFloat);

      tmpFloat = (knots[i])->GetY();
      yValues.Append(tmpFloat);

      tmpFloat = (knots[i])->GetZ();
      zValues.Append(tmpFloat);
   }

   CalculateDerivatives(xValues, this->firstDersX, this->secondDersX, this->thirdDersX);
   CalculateDerivatives(yValues, this->firstDersY, this->secondDersY, this->thirdDersY);
   CalculateDerivatives(zValues, this->firstDersZ, this->secondDersZ, this->thirdDersZ);
  
   
   // copy the knots into the spline
   for(i = 0 ; i < knots.Len() ; i++) {
      this->knots.Append(*knots.Index(i));
   }
}

CubicSpline::~CubicSpline(void) {
}

int CubicSpline::NumberOfKnots() {
   return this->knots.Len();
}


Pt3d CubicSpline::Evaluate(float t) {
   int tBefore;
   float x, y, z;
   
   // this should truncate t
   tBefore = (int)t;
   
   x = this->knots[tBefore].GetX() +
      (t - tBefore) * this->firstDersX[tBefore] +
      (((t - tBefore) * (t - tBefore)) / 2) * this->secondDersX[tBefore] +
      (((t - tBefore) * (t - tBefore) * (t - tBefore)) / 6) * this->thirdDersX[tBefore];

   y = this->knots[tBefore].GetY() +
      (t - tBefore) * this->firstDersY[tBefore] +
      (((t - tBefore) * (t - tBefore)) / 2) * this->secondDersY[tBefore] +
      (((t - tBefore) * (t - tBefore) * (t - tBefore)) / 6) * this->thirdDersY[tBefore];

   z = (this->knots[tBefore]).GetZ() +
      (t - tBefore) * this->firstDersZ[tBefore] +
      (((t - tBefore) * (t - tBefore)) / 2) * this->secondDersZ[tBefore] +
      (((t - tBefore) * (t - tBefore) * (t - tBefore)) / 6) * this->thirdDersZ[tBefore];

   return *(new Pt3d(x, y, z));
}
